import java.util.*;

public class BTreeIt<E>  {
    private class Node {
        E elem;
        Node l;
        Node r;
        Node p;

        void addLeft(Node n) {
            l = n;
            n.p = this;
        }
        void addRight(Node n) {
            r = n;
            n.p = this;
        }
        Node(E elem) { this.elem = elem; }
    }

    private Node root;

    private BTreeIt(Node n) {
        root = n;
    }

    private class PostOrderIterator implements Iterator<E> {
        Node next; 
        Node last;
 
        PostOrderIterator() {
            next = root;
            last = null;
            moveToLeftMostLeaf();
        }

        private void moveToLeftMostLeaf() {
            do {
                // go down left
                while (next.l != null) next = next.l;
                // maybe there is a node with no left but some right...
                // then go down rigth
                if (next.r != null) next = next.r;
            } while (next.l != null || next.r != null);
            // exit when both left and right are null
        } 

        public boolean hasNext() {
            if (next == null) return false;
            else return true;
        }

        public void remove() throws UnsupportedOperationException {
            throw new UnsupportedOperationException();
        }

        public E next() throws NoSuchElementException {
            if (next == null) throw new NoSuchElementException();
            E key = next.elem;
            // I already visited left and right, 
            // so I have to go up (and maybe right)
            last = next;
            next = next.p;
            if (next != null && last == next.l) {
                next = next.r;
                moveToLeftMostLeaf();
            }
            
            return key;
        }
    }

    /* ---------------------------------*/ 
    /* INTERFACE                        */
    /* ---------------------------------*/ 

    public BTreeIt(E elem) {
        root = new Node(elem);
    }

    public BTreeIt<E> addLeft(BTreeIt<E> t) {
        root.addLeft(t.root);
        return this;
    }

    public BTreeIt<E> addRight(BTreeIt<E> t) {
        root.addRight(t.root);
        return this;
    }

    public BTreeIt<E> linkToLeft(BTreeIt<E> t) {
        t.root.addLeft(root);
        return this;
    }

    public BTreeIt<E> linkToRight(BTreeIt t) {
        t.root.addRight(root);
        return this;
    }

    public BTreeIt<E> getLeftSubtree() {
        if (root == null) return null;
        else return new BTreeIt(root.l);
    }

    public BTreeIt<E> getRightSubtree() {
        if (root == null) return null;
        else return new BTreeIt(root.r);
    }


    Iterator<E> postOrderIterator() {
        return new PostOrderIterator();
    }

    /* ---------------------------------*/ 
    /* TEST                             */
    /* ---------------------------------*/ 
    public static void main(String args[]) {
        BTreeIt<String> mytree = new BTreeIt<String>("*");
        
        BTreeIt<String> ll = new BTreeIt<String>("+");

        BTreeIt<String> rr = new BTreeIt<String>("-");

        ll.addLeft(new BTreeIt<String>("2")).
            addRight(new BTreeIt<String>("-").
                     addRight(new BTreeIt<String>("-3"))).
            linkToLeft(mytree);
        rr.addLeft(new BTreeIt<String>("6")).
            addRight(new BTreeIt<String>("*").
                     addLeft(new BTreeIt<String>("2")).
                     addRight(new BTreeIt<String>("2"))
                     ).
            linkToRight(mytree);
        
        System.out.println("Post Order: ");
        Iterator<String> it = mytree.postOrderIterator();
        while (it.hasNext()) System.out.println(it.next());
    }  
}